﻿using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using MVR.FileManagementSecure;
using SimpleJSON;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

/* --- © 2022 JaxZoa <JaxZoa@gmail.com> --- */

namespace JaxPlugins
{
    public class CLS : MVRScript
    {
        public static string EXT_FULLBODY = "clp";
        public static string EXT_FULLTRIGGER = "clf";
        public static string EXT_DEST_TRIGGER = "cld";
        public static string MAIN_DIR;
        public static string FBP_DIR;
        public static string TP_DIR;
        public static string AUTOPRESETS_DIR;

        /* Meta Target Parameters */

        protected JSONStorableStringChooser js_positionTargetTransform;   
        protected JSONStorableStringChooser js_rotationTargetTransform;

        protected JSONStorableStringChooser js_GroundAtom;

        protected JSONStorableBool js_CanMove;
        protected JSONStorableStringChooser js_RotType;

        protected JSONStorableVector js_TargetPos;
        protected JSONStorableVector js_TargetRot;

        protected JSONStorableFloat js_DestinationReachedDelay;

        protected JSONStorableFloat js_MaxSpeed;
        protected JSONStorableFloat js_Acceleration;
        protected JSONStorableFloat js_RotationSpeed;
        protected JSONStorableFloat js_StopDistance;

        protected JSONStorableFloat js_RotationStopAngle;

        protected JSONStorableFloat js_RotationOffset;
        protected JSONStorableFloat js_RotationOffsetMoving;

        protected JSONStorableBool js_ResetPosWhenDisabled;
        protected JSONStorableBool js_UseStepTargets;

        protected JSONStorableBool js_UseInputsToMove;
        protected JSONStorableStringChooser js_InputSource;
        protected JSONStorableStringChooser js_MovementReferenceAtom;

        protected Atom groundAtom;
        private Transform targetTransform;
        protected Transform rotationTargetTransform;
        protected Transform positionTargetTransform;

        protected float tempRotationOffset = 0;
        protected Vector3 currentSpeed = Vector3.zero;
        protected List<Collider> groundColliders;

        protected float movementSpeedDirectional = 1;
        protected bool movementLock = false;

        protected Transform movementRefTransform;

        protected bool destinationReached = false;
        protected bool isIdle = true;
        public StateChanged idleStartDelegate;
        public Event destinationReachedDelegate;

        /* Actions parameters */


        protected float customInputX;
        protected float customInputY;

        protected float customInputX_Time, customInputY_Time;

        protected JSONStorableAction js_Action_MoveForward;
        protected JSONStorableAction js_Action_MoveBackward;
        protected JSONStorableAction js_Action_MoveLeft;
        protected JSONStorableAction js_Action_MoveRight;

        protected JSONStorableFloat js_AutoPresetIndex;
        protected JSONStorableFloat js_AutoPresetGroup;

        protected JSONStorableAction js_Action_SaveAutoPreset;
        protected JSONStorableAction js_Action_LoadAutoPreset;


        protected List<string> triggerList = new List<string> { "Select Trigger", "OnIdleStart", "OnStepIdleStart", "OnMovementStart", "OnDestinationReached"};
        protected List<string> transitionTriggerList = new List<string> { "Select Trigger", "Right Step", "Left Step", "Both Steps"};

        protected JSONStorableStringChooser js_TriggerEditorChooser;
        protected JSONStorableStringChooser js_TransitionTriggerEditorChooser;

        public EventTrigger trigger_OnIdleStart;
        public EventTrigger trigger_OnStepIdleStart;
        public EventTrigger trigger_OnMovementStart;
        public EventTrigger trigger_OnDestinationReached;

        public EventTrigger trigger_RightStep;
        public EventTrigger trigger_LeftStep;
        public EventTrigger trigger_BothSteps;

        protected List<EventTrigger> eventTriggers;
        protected List<EventTrigger> transitionTriggers;

        private float destinationReachedTime = 0;
        private bool destinationReachedTriggered = true;

        bool _idleState;
        bool _Initialized = false;

        public delegate void StateChanged(bool _newState);
        public delegate void Event();

        private float currentYPos = 0;

        public float CurrentYPos
        {
            get { return currentYPos; }
            set { currentYPos = value; }
        }

        /* Body Controller parameters */

        private BodyController bodyController;

        public override void Init()
        {
            /* Meta Target Init */

            MAIN_DIR = "Saves\\PluginData\\CLS2";
            FBP_DIR = MAIN_DIR + "\\GeneralPresets";
            TP_DIR = MAIN_DIR + "\\TriggerPresets";
            AUTOPRESETS_DIR = FBP_DIR + "\\AutoPresets";

            FileManagerSecure.CreateDirectory(FBP_DIR);
            FileManagerSecure.CreateDirectory(TP_DIR);
            FileManagerSecure.CreateDirectory(AUTOPRESETS_DIR);
            SuperController.singleton.fileBrowserUI.defaultPath = MAIN_DIR;
            SuperController.singleton.NormalizeMediaPath(MAIN_DIR + "\\");

            SimpleTriggerHandler.LoadAssets();
            InitTriggers();

            activeSliders = new List<UIDynamicSlider>();
            activeToggles = new List<UIDynamicToggle>();
            activeButtons = new List<UIDynamicButton>();
            activePopups = new List<UIDynamicPopup>();
            activeTextFields = new List<UIDynamicTextField>();
            activeSpacers = new List<UIDynamic>();

            js_TargetPos = new JSONStorableVector(this, "TargetPosition", Vector3.zero);
            js_TargetRot = new JSONStorableVector(this, "TargetRotation", Vector3.zero);
      
            js_positionTargetTransform = new JSONStorableStringChooser("Position Target Atom", GetAtomNames(), "[None]", "Pos Target Atom", Callback_TargetPositionAtomChanged);
            js_positionTargetTransform.popupOpenCallback = Callback_TargetAtomOpened;
            RegisterStringChooser(js_positionTargetTransform);

            js_rotationTargetTransform = new JSONStorableStringChooser("Rotation Target Atom", GetAtomNames(), "[None]", "Rot Target Atom", Callback_TargetRotationAtomChanged);
            js_rotationTargetTransform.popupOpenCallback = Callback_TargetAtomOpened;
            RegisterStringChooser(js_rotationTargetTransform);

            js_GroundAtom = new JSONStorableStringChooser("Ground Atom", GetAtomNames(), "[None]", "Ground Atom", Callback_GroundAtomChanged);
            js_GroundAtom.popupOpenCallback = Callback_GroundAtomOpened;
            RegisterStringChooser(js_GroundAtom);

            js_TriggerEditorChooser = new JSONStorableStringChooser("TriggerEditorChooser", triggerList, triggerList[0], "Trigger Editor", Callback_TriggerEditorChooser);
            RegisterStringChooser(js_TriggerEditorChooser);

            js_TransitionTriggerEditorChooser = new JSONStorableStringChooser("TransitionTriggerEditorChooser", transitionTriggerList, transitionTriggerList[0], "Transition Trigger Editor", Callback_TransitionTriggerEditorChooser);
            RegisterStringChooser(js_TransitionTriggerEditorChooser);

            js_CanMove = new JSONStorableBool("Can Move", true);
            RegisterBool(js_CanMove);
            
            js_MaxSpeed = new JSONStorableFloat("Max Speed", 0.6f, 0.0f, 1.5f, true);
            RegisterFloat(js_MaxSpeed);
            
            js_Acceleration = new JSONStorableFloat("Acceleration", 1.5f, 0.2f, 5f, true);
            RegisterFloat(js_Acceleration);

            js_StopDistance = new JSONStorableFloat("Stop Distance", 0.25f, 0.01f, 1f, false);
            RegisterFloat(js_StopDistance);

            js_RotType = new JSONStorableStringChooser("RotationType", new List<string> { "No Rot", "Face Target", "Face Rot Target Idle", "Copy Target", "Copy Target Idle" }, "Face Target", "Rotation Type");
            RegisterStringChooser(js_RotType);            

            js_RotationSpeed = new JSONStorableFloat("Rotation Speed", 1f, 0.1f, 10f, true);
            RegisterFloat(js_RotationSpeed);

            js_RotationOffset = new JSONStorableFloat("Rotation Offset Idle", 0, -180, 180);
            RegisterFloat(js_RotationOffset);

            js_RotationOffsetMoving = new JSONStorableFloat("Rotation Offset Moving", 0, -180, 180);
            RegisterFloat(js_RotationOffsetMoving);

            js_RotationStopAngle = new JSONStorableFloat("Rotation Stop Angle", 15f, 0f, 45f);
            RegisterFloat(js_RotationStopAngle);

            js_DestinationReachedDelay = new JSONStorableFloat("Destination Reached Delay", 2f, 0.1f, 10f);
            RegisterFloat(js_DestinationReachedDelay);

            js_ResetPosWhenDisabled = new JSONStorableBool("ResetPosWhenDisabled", true);
            RegisterBool(js_ResetPosWhenDisabled);

            js_UseStepTargets = new JSONStorableBool("UseStepTargets", false);
            RegisterBool(js_UseStepTargets);

            js_UseInputsToMove = new JSONStorableBool("Use Inputs To Move", false);
            RegisterBool(js_UseInputsToMove);

            js_InputSource = new JSONStorableStringChooser("InputSource", new List<string> { "WASD", "ZQSD", "Controller_DPad", "Controller_LeftStick", "Custom" }, "WASD", "Input Source");
            RegisterStringChooser(js_InputSource);

            js_MovementReferenceAtom = new JSONStorableStringChooser("MovementRefAtom", new List<string>(), "Select Atom", "Movement Reference", Callback_MovementInputReferenceChanged);
            js_MovementReferenceAtom.popupOpenCallback = Callback_MovementInputReferenceOpened;
            RegisterStringChooser(js_MovementReferenceAtom);

            groundColliders = new List<Collider>();

            Callback_RefreshStepTargetList();

            js_Action_MoveForward = new JSONStorableAction("Move Forward", Callback_MoveForward);
            RegisterAction(js_Action_MoveForward);

            js_Action_MoveBackward = new JSONStorableAction("Move Backward", Callback_MoveBackward);
            RegisterAction(js_Action_MoveBackward);

            js_Action_MoveLeft = new JSONStorableAction("Move Left", Callback_MoveLeft);
            RegisterAction(js_Action_MoveLeft);

            js_Action_MoveRight = new JSONStorableAction("Move Right", Callback_MoveRight);
            RegisterAction(js_Action_MoveRight);

            js_Action_SaveAutoPreset = new JSONStorableAction("Save Auto Preset", Callback_SaveAutoPreset);
            RegisterAction(js_Action_SaveAutoPreset);

            js_Action_LoadAutoPreset = new JSONStorableAction("Load Auto Preset", Callback_LoadAutoPreset);
            RegisterAction(js_Action_LoadAutoPreset);

            js_AutoPresetGroup = new JSONStorableFloat("AutoPreset Group", 0, Callback_AutoPresetGroupChanged, 0, 1000, true);
            RegisterFloat(js_AutoPresetGroup);

            js_AutoPresetIndex = new JSONStorableFloat("AutoPreset Index", 0, Callback_AutoPresetIndexChanged, 0, 1000, true);
            RegisterFloat(js_AutoPresetIndex);

            /* Body Controller init */

            bodyController = new BodyController_Complete(this);

            SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;

            CreateGlobalUI();
        }

        public void InitTriggers()
        {
            eventTriggers = new List<EventTrigger>();
            transitionTriggers = new List<EventTrigger>();

            trigger_OnIdleStart = new EventTrigger(this, "OnIdleStart");
            trigger_OnStepIdleStart = new EventTrigger(this, "OnStepIdleStart");
            trigger_OnMovementStart = new EventTrigger(this, "OnMovementStart");
            trigger_OnDestinationReached = new EventTrigger(this, "OnDestinationReached");

            eventTriggers.Add(trigger_OnIdleStart);
            eventTriggers.Add(trigger_OnStepIdleStart);
            eventTriggers.Add(trigger_OnMovementStart);
            eventTriggers.Add(trigger_OnDestinationReached);

            idleStartDelegate += Event_StartIdle;
            destinationReachedDelegate += Event_DestinationReached;

            trigger_RightStep = new EventTrigger(this, "RightStep");
            trigger_LeftStep = new EventTrigger(this, "LeftStep");
            trigger_BothSteps = new EventTrigger(this, "BothSteps");

            transitionTriggers.Add(trigger_RightStep);
            transitionTriggers.Add(trigger_LeftStep);
            transitionTriggers.Add(trigger_BothSteps);
        }

        void Start()
        {
            /* Meta Target Start*/

            GameObject _targetDummy = new GameObject("TargetDummy");

            targetTransform = _targetDummy.transform;

            FreeControllerV3 hipController = GetControllerByName("hipControl");

            if (hipController == null)
                SuperController.LogError("You need to put this script on a Person atom !");

            if (js_TargetPos.val == Vector3.zero && js_TargetRot.val == Vector3.zero)
            {
                js_TargetPos.val = hipController.transform.position;
                js_TargetPos.y = 0;

                js_TargetRot.y = hipController.transform.rotation.eulerAngles.y;
            }

            targetTransform.position = js_TargetPos.val;
            targetTransform.rotation = Quaternion.Euler(js_TargetRot.val);

            bodyController.Start();

            _Initialized = true;
        }

        void Update()
        {           
            try
            {
                _idleState = isIdle;

                ComputeMovementInputs();
                ComputeMetaTargetPosition();
                ComputeMetaTargetRotation();

                if (_idleState != isIdle)
                    idleStartDelegate.Invoke(isIdle);

                _idleState = bodyController.StepCycle.IsIdle;

                bodyController.Update();

                UpdateTransitionTriggers();

                if(!bodyController.StepCycle.IsIdle && bodyController.StepCycle.IsIdle != _idleState)
                {
                    trigger_OnMovementStart.Trigger();
                }

                if (destinationReached && (Time.time - destinationReachedTime) > js_DestinationReachedDelay.val && !destinationReachedTriggered)
                {
                    destinationReachedDelegate.Invoke();

                    destinationReachedTriggered = true;
                }
            } 
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }            
        }

        void OnDestroy()
        {
            GameObject.Destroy(targetTransform.gameObject);

            bodyController.OnDestroy();

            SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;

            foreach (EventTrigger et in eventTriggers)
            {
                et.Remove();
            }
        }

        Vector3 _targetAtomPos;
        Vector3 _deltaTarget;
        float _rotAmount;
        Quaternion _atomTargetRot;
        Transform _currentRotTargetTransform;
        Quaternion _targetRot;
        Quaternion _rotToTarget;
        float _angleToFaceTarget;

        Vector3 _inputDrivenTargetPos;
        float _xAxis, _yAxis;
        Vector3 _tmpDir, _tmpPos;

        void ComputeMetaTargetPosition()
        {
            if (positionTargetTransform == null || !js_CanMove.val || movementLock)
            {
                isIdle = true;
                currentSpeed = Vector3.zero;

                return;
            }

            _targetAtomPos = positionTargetTransform.position;
            _targetAtomPos.y = 0;

            _deltaTarget = _targetAtomPos - targetTransform.position;

            if (_deltaTarget.magnitude > js_StopDistance.val)
            {
                currentSpeed += _deltaTarget.normalized * js_Acceleration.val * Time.deltaTime;
                currentSpeed = Vector3.ClampMagnitude(currentSpeed, js_MaxSpeed.val);

                isIdle = false;

                destinationReached = false;
                destinationReachedTriggered = false;
            }
            else if (_deltaTarget.magnitude - currentSpeed.magnitude > js_StopDistance.val)
            {
                currentSpeed -= _deltaTarget.normalized * js_Acceleration.val * Time.deltaTime;
                currentSpeed = Vector3.ClampMagnitude(currentSpeed, js_MaxSpeed.val);

                isIdle = false;

                destinationReached = false;
                destinationReachedTriggered = false;
            }
            else
            {
                currentSpeed = Vector3.zero;

                isIdle = true;

                if (!destinationReached)
                {
                    destinationReached = true;

                    destinationReachedTime = Time.time;
                }
            }

            targetTransform.position = targetTransform.position + (currentSpeed * Time.deltaTime);

            js_TargetPos.val = targetTransform.position;
        }       

        void ComputeMetaTargetRotation()
        {
            movementSpeedDirectional = 1;

            if (js_RotType.val == "No Rot")
                return;

            if (positionTargetTransform == null && rotationTargetTransform == null)
                return;

            _rotAmount = js_RotationSpeed.val * Time.deltaTime * 100;

            if (rotationTargetTransform != null)
                _currentRotTargetTransform = rotationTargetTransform.transform;
            else
                _currentRotTargetTransform = positionTargetTransform.transform;

            if (_currentRotTargetTransform.up.y > 0)
                _atomTargetRot = Quaternion.Euler(0, _currentRotTargetTransform.rotation.eulerAngles.y, 0);
            else
                _atomTargetRot = Quaternion.Euler(0, _currentRotTargetTransform.rotation.eulerAngles.y - 180, 0);

            _deltaTarget = _currentRotTargetTransform.position - targetTransform.position;
            _deltaTarget.y = 0;
            _deltaTarget.Normalize();

            _targetRot = targetTransform.rotation;            

            switch (js_RotType.val)
            {
                case "Face Target":
                    if(isIdle)
                        _rotToTarget = Quaternion.LookRotation(_deltaTarget, Vector3.up) * Quaternion.Euler(0, js_RotationOffset.val + tempRotationOffset, 0);
                    else
                        _rotToTarget = Quaternion.LookRotation(_deltaTarget, Vector3.up) * Quaternion.Euler(0, js_RotationOffsetMoving.val + tempRotationOffset, 0);

                    _targetRot = Quaternion.RotateTowards(_targetRot, _rotToTarget, _rotAmount);
                    break;

                case "Face Rot Target Idle":

                    if (rotationTargetTransform == null || positionTargetTransform == null)
                        return;

                    if (isIdle)
                    {
                        _deltaTarget = rotationTargetTransform.position - targetTransform.position;
                        _deltaTarget.y = 0;
                        _deltaTarget.Normalize();

                        _rotToTarget = Quaternion.LookRotation(_deltaTarget, Vector3.up) * Quaternion.Euler(0, js_RotationOffset.val + tempRotationOffset, 0);
                        _targetRot = Quaternion.RotateTowards(_targetRot, _rotToTarget, _rotAmount);
                    }
                    else
                    {
                        _deltaTarget = positionTargetTransform.position - targetTransform.position;
                        _deltaTarget.y = 0;
                        _deltaTarget.Normalize();

                        _rotToTarget = Quaternion.LookRotation(_deltaTarget, Vector3.up) * Quaternion.Euler(0, js_RotationOffsetMoving.val + tempRotationOffset, 0);
                        _targetRot = Quaternion.RotateTowards(_targetRot, _rotToTarget, _rotAmount);
                    }
                    break;
                case "Copy Target":
                    if (isIdle)
                    {
                        _targetRot = Quaternion.RotateTowards(_targetRot, _atomTargetRot * Quaternion.Euler(0, js_RotationOffset.val + tempRotationOffset, 0), _rotAmount);
                    }
                    else
                    {
                        _targetRot = Quaternion.RotateTowards(_targetRot, _atomTargetRot * Quaternion.Euler(0, js_RotationOffsetMoving.val + tempRotationOffset, 0), _rotAmount);
                    }
                    break;
                case "Copy Target Idle":
                    if (isIdle)
                    {
                        _rotToTarget = Quaternion.LookRotation(_deltaTarget, Vector3.up) * Quaternion.Euler(0, js_RotationOffset.val + tempRotationOffset, 0);
                        _targetRot = Quaternion.RotateTowards(_targetRot, _atomTargetRot * Quaternion.Euler(0, js_RotationOffset.val + tempRotationOffset, 0), _rotAmount);
                    }
                    else
                    {
                        _rotToTarget = Quaternion.LookRotation(_deltaTarget, Vector3.up) * Quaternion.Euler(0, js_RotationOffsetMoving.val + tempRotationOffset, 0);
                        _targetRot = Quaternion.RotateTowards(_targetRot, _rotToTarget * Quaternion.Euler(0, js_RotationOffsetMoving.val + tempRotationOffset, 0), _rotAmount);
                    }
                    break;
            }

            targetTransform.rotation = _targetRot;

            _angleToFaceTarget = Vector3.SignedAngle(targetTransform.forward, _deltaTarget.normalized, Vector3.up);

            if (!isIdle &&(js_RotType.val == "Face Target" || js_RotType.val == "Copy Target Idle" || js_RotType.val == "Face Rot Target Idle"))
            {
                if (Mathf.Abs(_angleToFaceTarget) > js_RotationStopAngle.val + Mathf.Abs(js_RotationOffsetMoving.val))
                {
                    currentSpeed = Vector3.zero;
                    isIdle = true;
                }
            }

            movementSpeedDirectional = Vector3.Dot(targetTransform.forward, _deltaTarget);
            movementSpeedDirectional = Mathf.Abs(movementSpeedDirectional);

            js_TargetRot.val = targetTransform.rotation.eulerAngles;
        }

        void ComputeMovementInputs()
        {
            _xAxis = 0;
            _yAxis = 0;

            if (!js_UseInputsToMove.val || positionTargetTransform == null || movementRefTransform == null)
                return;

            switch(js_InputSource.val)
            {
                case "WASD":  

                    if (Input.GetKey(KeyCode.W))
                        _yAxis = 1;
                    else if (Input.GetKey(KeyCode.S))
                        _yAxis = -1;

                    if (Input.GetKey(KeyCode.D))
                        _xAxis = 1;
                    else if (Input.GetKey(KeyCode.A))
                        _xAxis = -1;

                    break;

                case "ZQSD":

                    if (Input.GetKey(KeyCode.Z))
                        _yAxis = 1;
                    else if (Input.GetKey(KeyCode.S))
                        _yAxis = -1;

                    if (Input.GetKey(KeyCode.D))
                        _xAxis = 1;
                    else if (Input.GetKey(KeyCode.Q))
                        _xAxis = -1;

                    break;

                case "Controller_DPad":

                    _yAxis = Input.GetAxis(JoystickControl.singleton.x360DPadYAxisName);
                    _xAxis = Input.GetAxis(JoystickControl.singleton.x360DPadXAxisName);

                    break; 

                case "Controller_LeftStick":

                    _yAxis = -Input.GetAxis(JoystickControl.singleton.leftStickYAxisName);
                    _xAxis = Input.GetAxis(JoystickControl.singleton.leftStickXAxisName);

                    break;

                case "Custom":

                    if (Time.time > customInputX_Time + 0.1f)
                        customInputX = 0;

                    if (Time.time > customInputY_Time + 0.1f)
                        customInputY = 0;

                    _yAxis = customInputY;
                    _xAxis = customInputX;

                    break;
            }

            _tmpPos = positionTargetTransform.position;
            _tmpPos.y = GetMainController().transform.position.y;

            positionTargetTransform.position = _tmpPos;

            _inputDrivenTargetPos = positionTargetTransform.position;

            _tmpDir = movementRefTransform.forward;
            _tmpDir.y = 0;
            _tmpDir.Normalize();

            _inputDrivenTargetPos += _tmpDir * _yAxis * Time.deltaTime * (1 + js_MaxSpeed.val * 2);

            _tmpDir = movementRefTransform.right;
            _tmpDir.y = 0;
            _tmpDir.Normalize();

            _inputDrivenTargetPos += _tmpDir * _xAxis * Time.deltaTime * (1 + js_MaxSpeed.val * 2);

            if ((_inputDrivenTargetPos - GetMainController().transform.position).magnitude > 0.3f)
                return;

            positionTargetTransform.position = _inputDrivenTargetPos;
        }

        void UpdateTransitionTriggers()
        {
            trigger_RightStep.TriggerTransition(bodyController.StepCycle.GetStep(1).GetStepStrength(0.5f));
            trigger_LeftStep.TriggerTransition(bodyController.StepCycle.GetStep(3).GetStepStrength(0.5f));
            trigger_BothSteps.TriggerTransition(bodyController.StepCycle.GetStep(1).GetStepStrength(0.5f) + bodyController.StepCycle.GetStep(3).GetStepStrength(0.5f));
        }

        /* Getters & Setters */

        public Transform TargetTransform
        {
            get
            {
                return targetTransform;
            }
        }

        public Transform PositionTargetTransform
        {
            get
            {
                return positionTargetTransform;
            }
        }

        public Transform RotationTargetTransform
        {
            get
            {
                if (rotationTargetTransform != null)
                    return RotationTargetTransform;
                else
                    return PositionTargetTransform;
            }
        }

        public BodyController BodyController
        {
            get
            {
                return bodyController;
            }
        }

        public float GetTargetBodyHeight
        {
            get
            {
                if (positionTargetTransform != null)
                    return positionTargetTransform.position.y - 0.5f;
                else
                    return 1;
            }
        }

        public Vector3 CurrentSpeed
        {
            get
            {
                return currentSpeed;
            }
        }

        public float SpeedFactor
        {
            get { return currentSpeed.magnitude / js_MaxSpeed.val; }
        }

        public float MovementSpeedDirectional
        {
            get
            {
                return movementSpeedDirectional;
            }
        }

        public float RotationOffset
        {
            get { return js_RotationOffset.val; }
            set { js_RotationOffset.val = value; }
        }

        public float RotationOffsetMoving
        {
            get { return js_RotationOffsetMoving.val; }
            set { js_RotationOffsetMoving.val = value; }
        }

        public float TempRotationOffset
        {
            get { return tempRotationOffset; }
            set { tempRotationOffset = value; }
        }

        public bool IsIdle
        {
            get
            {
                return isIdle;
            }
        }

        public List<Collider> GroundColliders
        {
            get
            {               
                return groundColliders;
            }
        }

        public bool UseStepTargets
        {
            get { return js_UseStepTargets.val; }
        }

        public List<MVRScript> StepTargets
        {
            get { return stepTargets; }
        }

        public float CustomInputX
        {
            get { return customInputX; }

            set { customInputX = value; }
        }

        public float CustomInputY
        {
            get { return customInputX; }

            set { customInputX = value; }
        }


        /* Movement Functions */

        public void LockMovement()
        {
            movementLock = true;
        }

        public void UnlockMovement()
        {
            movementLock = false;
        }

        void OnEnable()
        {
            if(js_ResetPosWhenDisabled.val && _Initialized)
            {
                Vector3 _tarPos = bodyController.CurrentBodyPos;
                _tarPos.y = 0;

                targetTransform.position = _tarPos;
                targetTransform.rotation = bodyController.CurrentBodyRot;

                bodyController.StepCycle.ResetSteps();
            }
        }

        void OnDisable()
        {
            
        }

        /* UI */

        protected List<UIDynamicSlider> activeSliders;
        protected List<UIDynamicToggle> activeToggles;
        protected List<UIDynamicButton> activeButtons;
        protected List<UIDynamicPopup> activePopups;
        protected List<UIDynamicTextField> activeTextFields;
        protected List<UIDynamic> activeSpacers;

        protected string activeUIPanel = "None";

        void CreateGlobalUI()
        {
            CreateMainButton("Presets", CreatePresetsUI);
            CreateMainButton("Triggers", CreateTriggersUI);

            CreateStaticText("Title", "<color=#000><size=35><b>Global</b></size></color>", false, 50);

            CreateMainButton("Movement/Rotation", CreateMovementUI);

            CreateMainButton("Step Targets", CreateStepsTargetUI);

            CreateMainButton("Input Control", CreateMovementInputsUI);

            bodyController.CreateGlobalUI();

            CreateMovementUI();
        }

        public void CreateMovementUI()
        {
            ClearUI();

            activeUIPanel = "Movement";

            CreateTempStaticText("Title", "<color=#000><size=35><b>Targets</b></size></color>", 40);

            CreateTempFilterablePopup(js_positionTargetTransform);
            CreateTempFilterablePopup(js_rotationTargetTransform);
            CreateTempFilterablePopup(js_GroundAtom);

            CreateTempStaticText("Title", "<color=#000><size=35><b>Movement</b></size></color>", 40);

            CreateTempToggle(js_CanMove);
            CreateTempSlider(js_MaxSpeed);
            CreateTempSlider(js_Acceleration);
            CreateTempSlider(js_StopDistance);

            CreateTempSpacer(15f);

            CreateTempToggle(js_ResetPosWhenDisabled);

            CreateTempStaticText("Title", "<color=#000><size=35><b>Rotation</b></size></color>", 40);

            CreateTempFilterablePopup(js_RotType);
            CreateTempSlider(js_RotationSpeed);
            CreateTempSlider(js_RotationStopAngle);
            CreateTempSlider(js_RotationOffset);
            CreateTempSlider(js_RotationOffsetMoving);
        }

        public void CreatePresetsUI()
        {
            ClearUI();

            activeUIPanel = "Presets";

            CreateTempStaticText("Title", "<color=#000><size=35><b>Presets</b></size></color>", 50);

            CreateTempButton("Save Preset", Callback_CreateNewFullBodyPreset);

            CreateTempButton("Load Preset", Callback_LoadFullBodyPreset);

            CreateTempSlider(js_AutoPresetIndex);
            CreateTempSlider(js_AutoPresetGroup);

            bodyController.BodyControllerPresetsUI();
        }

        public void CreateTriggersUI()
        {
            ClearUI();

            activeUIPanel = "Triggers";

            CreateTempStaticText("Title", "<color=#000><size=35><b>Triggers</b></size></color>", 50);

            CreateTempPopup(js_TriggerEditorChooser);

            CreateTempButton("Save Full Trigger", Callback_Trigger_SaveFullTriggerPreset);

            CreateTempButton("Load Full Trigger", Callback_Trigger_LoadFullTriggerPreset);

            CreateTempButton("Save Destination Reached", Callback_Trigger_SaveDestinationReachedPreset);

            CreateTempButton("Load Destination Reached", Callback_Trigger_LoadDestinationReachedPreset);

            CreateTempSpacer(25f);

            CreateTempSlider(js_DestinationReachedDelay);

            bodyController.BodyControllerTriggersUI();

            CreateTempSpacer(25f);

            CreateTempPopup(js_TransitionTriggerEditorChooser);
        }

        public void CreateStepsTargetUI()
        {
            ClearUI();

            activeUIPanel = "StepTarget";

            CreateTempStaticText("Title", "<color=#000><size=35><b>Step Targets</b></size></color>", 50);

            CreateTempToggle(js_UseStepTargets);

            CreateTempButton("Refresh Step Targets List", Callback_RefreshStepTargetList);
        }

        public void CreateMovementInputsUI()
        {
            ClearUI();

            activeUIPanel = "MovementInputs";

            CreateTempStaticText("Title", "<color=#000><size=35><b>Input Control</b></size></color>", 40);

            CreateTempToggle(js_UseInputsToMove);
            CreateTempPopup(js_InputSource);
            CreateTempFilterablePopup(js_MovementReferenceAtom);

            bodyController.BodyControllerMovementUI();
        }

        public string ActiveUIPanel
        {
            get { return activeUIPanel; }
            set { activeUIPanel = value; }
        }

        /* UI Helpers */

        public void CreateMainButton(string _text, UnityEngine.Events.UnityAction _callback)
        {
            ColorBlock _colors = new ColorBlock();
            _colors.normalColor = new Color(0.3f, 0.5f, 0.7f);
            _colors.pressedColor = new Color(0.2f, 0.4f, 0.6f);
            _colors.highlightedColor = new Color(0.1f, 0.1f, 0.3f);
            _colors.colorMultiplier = 1f;

            UIDynamicButton _btn = CreateButton(_text, false);
            _btn.button.onClick.AddListener(_callback);
            _btn.button.colors = _colors;

            _btn.textColor = new Color(0.9f, 0.9f, 0.9f);
            _btn.buttonText.fontSize = 35;
        }

        public void CreateTempSlider(JSONStorableFloat _data)
        {
            activeSliders.Add(CreateSlider(_data, true));
        }

        public void CreateTempToggle(JSONStorableBool _data)
        {
            activeToggles.Add(CreateToggle(_data, true));
        }

        public void CreateTempButton(string _text, UnityEngine.Events.UnityAction _callback)
        {
            UIDynamicButton _btn = CreateButton(_text, true);
            _btn.button.onClick.AddListener(_callback);

            activeButtons.Add(_btn);
        }

        public void CreateTempSpacer(float _height)
        {
            UIDynamic _spacer = CreateSpacer(true);
            _spacer.height = _height;
            activeSpacers.Add(_spacer);
        }

        public void CreateTempPopup(JSONStorableStringChooser _data)
        {
            activePopups.Add(CreatePopup(_data, true));
        }

        public void CreateTempFilterablePopup(JSONStorableStringChooser _data)
        {
            activePopups.Add(CreateFilterablePopup(_data, true));
        }

        public UIDynamicTextField CreateStaticText(string DescTitle, string DescText, bool rightSide, int fieldHeight)
        {
            JSONStorableString staticDescString = new JSONStorableString(DescTitle, DescText);
            UIDynamicTextField staticDescStringField = CreateTextField(staticDescString, rightSide);
            staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
            LayoutElement sdsfLayout = staticDescStringField.GetComponent<LayoutElement>();
            sdsfLayout.preferredHeight = sdsfLayout.minHeight = fieldHeight;
            staticDescStringField.height = fieldHeight;
            DisableScrollOnText(staticDescStringField);

            return staticDescStringField;
        }

        public UIDynamicTextField CreateTempStaticText(string DescTitle, string DescText, int fieldHeight)
        {
            activeTextFields.Add(CreateStaticText(DescTitle, DescText, true, fieldHeight));

            return activeTextFields[activeTextFields.Count - 1];
        }


        public void ClearUI()
        {
            for (int i = 0; i < activeSliders.Count; i++)
            {
                RemoveSlider(activeSliders[i]);
            }

            for (int i = 0; i < activeToggles.Count; i++)
            {
                RemoveToggle(activeToggles[i]);
            }

            for (int i = 0; i < activeButtons.Count; i++)
            {
                RemoveButton(activeButtons[i]);
            }

            for (int i = 0; i < activeSpacers.Count; i++)
            {
                RemoveSpacer(activeSpacers[i]);
            }

            for (int i = 0; i < activePopups.Count; i++)
            {
                RemovePopup(activePopups[i]);
            }
            
            for (int i = 0; i < activeTextFields.Count; i++)
            {
                RemoveTextField(activeTextFields[i]);
            }

            activeTextFields = new List<UIDynamicTextField>();
            activeSpacers = new List<UIDynamic>();
            activeButtons = new List<UIDynamicButton>();
            activeSliders = new List<UIDynamicSlider>();
            activePopups = new List<UIDynamicPopup>();
            activeToggles = new List<UIDynamicToggle>();
        }

        /* Callbacks */

        public void Callback_TargetPositionAtomChanged(string s)
        {
            List<Atom> atoms = GetSceneAtoms();

            if (s == "[None]")
            {
                positionTargetTransform = null;
            }
            else if(s == "[Player]")
            {
                positionTargetTransform = CameraTarget.centerTarget.transform;
            }
            else
            {
                foreach (Atom a in atoms)
                {
                    if (a.name == s)
                    {
                        positionTargetTransform = a.mainController.transform;
                    }
                }
            }

        }

        public void Callback_TargetRotationAtomChanged(string s)
        {
            List<Atom> atoms = GetSceneAtoms();

            if (s == "[None]")
            {
                rotationTargetTransform = null;
            }
            else if (s == "[Player]")
            {
                rotationTargetTransform = CameraTarget.centerTarget.transform;
            }
            else
            {
                foreach (Atom a in atoms)
                {
                    if (a.name == s)
                    {
                        rotationTargetTransform = a.mainController.transform;
                    }
                }
            }

        }

        public void Callback_TargetAtomOpened()
        {
            js_positionTargetTransform.choices = GetAtomNames();
            js_rotationTargetTransform.choices = js_positionTargetTransform.choices;
        }

        public void Callback_GroundAtomChanged(string s)
        {
            List<Atom> atoms = GetSceneAtoms();

            if (s == "[None]")
            {
                groundAtom = null;
            }
            else
            {
                foreach (Atom a in atoms)
                {
                    if (a.name == s)
                    {
                        groundAtom = a;
                        groundColliders = new List<Collider>();
                        groundColliders.AddRange(groundAtom.GetComponentsInChildren<Collider>());
                    }
                }
            }
        }

        public void Callback_GroundAtomOpened()
        {
            List<string> _choices = GetAtomNames();
            _choices.Remove("[Player]");

            js_GroundAtom.choices = _choices;
        }

        /* Actions Callbacks */

        public void Callback_MoveForward()
        {
            customInputY_Time = Time.time;
            customInputY = 1;
        }

        public void Callback_MoveBackward()
        {
            customInputY_Time = Time.time;
            customInputY = -1;
        }

        public void Callback_MoveLeft()
        {
            customInputX_Time = Time.time;
            customInputX = -1;
        }

        public void Callback_MoveRight()
        {
            customInputX_Time = Time.time;
            customInputX = 1;
        }

        public void Callback_SaveAutoPreset()
        {
            SaveFullBodyPreset(AUTOPRESETS_DIR + "\\Preset" + Mathf.FloorToInt(js_AutoPresetGroup.val).ToString() + "_" + Mathf.FloorToInt(js_AutoPresetIndex.val).ToString() + ".clp");
        }

        public void Callback_LoadAutoPreset()
        {
            LoadFullBodyPreset(AUTOPRESETS_DIR + "\\Preset" + Mathf.FloorToInt(js_AutoPresetGroup.val).ToString() + "_" + Mathf.FloorToInt(js_AutoPresetIndex.val).ToString() + ".clp");
        }

        public void Callback_AutoPresetIndexChanged(float _val)
        {
            js_AutoPresetIndex.val = Mathf.FloorToInt(js_AutoPresetIndex.val);
        }

        public void Callback_AutoPresetGroupChanged(float _val)
        {
            js_AutoPresetGroup.val = Mathf.FloorToInt(js_AutoPresetGroup.val);
        }

        /* Full body presets */

        public void Callback_CreateNewFullBodyPreset()
        {
            // Update the browser to be a Save browser
            uFileBrowser.FileBrowser browser = SuperController.singleton.fileBrowserUI;

            GenerateShortcuts();

            browser.SetTitle("Save CLS Preset");
            browser.fileRemovePrefix = null;
            browser.hideExtension = false;
            browser.keepOpen = false;
            browser.fileFormat = EXT_FULLBODY;
            browser.showDirs = true;
            browser.browseVarFilesAsDirectories = false;
            browser.SetTextEntry(true);
            browser.ActivateFileNameField();
            browser.clearCurrentPathOnHide = false;
            browser.fileEntryField.text = String.Format("{0}.{1}", "", EXT_FULLBODY);
            browser.Show(SaveFullBodyPreset);
        }

        public void Callback_LoadFullBodyPreset()
        {
            uFileBrowser.FileBrowser browser = SuperController.singleton.fileBrowserUI;

            GenerateShortcuts();

            browser.SetTitle("Load CLS Preset");
            browser.fileRemovePrefix = null;
            browser.hideExtension = false;
            browser.keepOpen = false;
            browser.fileFormat = EXT_FULLBODY;
            browser.showDirs = true;
            browser.browseVarFilesAsDirectories = true;
            browser.SetTextEntry(false);
            browser.ActivateFileNameField();
            browser.clearCurrentPathOnHide = false;
            browser.Show(LoadFullBodyPreset);
        }

        public void SaveFullBodyPreset(string _path)
        {
            if (_path == "")
                return;

            JSONClass _saveJSON = new JSONClass();

            if (!_path.ToLower().EndsWith(EXT_FULLBODY.ToLower()))
            {
                _path += "." + EXT_FULLBODY;
            }

            string[] _pathSplit = _path.Split('\\');
            string _dirPath = "";

            for(int i = 0; i < _pathSplit.Length - 1; i++)
            {
                _dirPath += _pathSplit[i] + "\\";
            }

            SuperController.singleton.fileBrowserUI.defaultPath = _dirPath;

            try
            {
                _saveJSON = bodyController.GetAllBodyParameters();
            }
            catch (Exception e)
            {
                SuperController.LogError("Exception caught: " + e);
            }

            _saveJSON["MovementParameters"] = new JSONArray();

            _saveJSON["MovementParameters"][0] = js_Acceleration.val.ToString();
            _saveJSON["MovementParameters"][1] = js_CanMove.val.ToString();
            _saveJSON["MovementParameters"][2] = js_MaxSpeed.val.ToString();
            _saveJSON["MovementParameters"][3] = js_RotationSpeed.val.ToString();
            _saveJSON["MovementParameters"][4] = js_RotType.val.ToString();
            _saveJSON["MovementParameters"][5] = js_StopDistance.val.ToString();
            _saveJSON["MovementParameters"][6] = js_positionTargetTransform.val.ToString();
            _saveJSON["MovementParameters"][7] = js_RotationOffset.val.ToString();
            _saveJSON["MovementParameters"][8] = js_DestinationReachedDelay.val.ToString();
            _saveJSON["MovementParameters"][9] = js_ResetPosWhenDisabled.val.ToString();
            _saveJSON["MovementParameters"][10] = js_UseStepTargets.val.ToString();
            _saveJSON["MovementParameters"][11] = js_RotationStopAngle.val.ToString();

            _saveJSON["TransitionTriggers"] = new JSONArray();

            _saveJSON["TransitionTriggers"][0] = trigger_RightStep.GetJSON().AsObject;
            _saveJSON["TransitionTriggers"][1] = trigger_LeftStep.GetJSON().AsObject;
            _saveJSON["TransitionTriggers"][2] = trigger_BothSteps.GetJSON().AsObject;

            SaveJSON(_saveJSON, _path);
        }

        public void LoadFullBodyPreset(string _path)
        {
            if (_path == "")
                return;

            if (!_path.ToLower().EndsWith(EXT_FULLBODY.ToLower()))
            {
                SuperController.LogError("CLS: Tried to load a preset file of the wrong type");
                return;
            }

            JSONNode _jsonFile = LoadJSON(_path);

            string[] _pathSplit = _path.Split('/');
            string _dirPath = "";

            for (int i = 0; i < _pathSplit.Length - 1; i++)
            {
                _dirPath += _pathSplit[i] + "\\";
            }

            SuperController.singleton.fileBrowserUI.defaultPath = _dirPath;

            bodyController.SetAllParametersFromSaveFile(_jsonFile);

            js_Acceleration.val = _jsonFile["MovementParameters"][0].AsFloat;
            js_CanMove.val = _jsonFile["MovementParameters"][1].AsBool;
            js_MaxSpeed.val = _jsonFile["MovementParameters"][2].AsFloat;
            js_RotationSpeed.val = _jsonFile["MovementParameters"][3].AsFloat;
            js_RotType.val = _jsonFile["MovementParameters"][4];
            js_StopDistance.val = _jsonFile["MovementParameters"][5].AsFloat;
            //js_TargetAtom.val = _jsonFile["MovementParameters"][6];
            //js_RotationOffset.val = _jsonFile["MovementParameters"][7].AsFloat;
            js_DestinationReachedDelay.val = _jsonFile["MovementParameters"][8].AsFloat;
            //js_ResetPosWhenDisabled.val = _jsonFile["MovementParameters"][9].AsBool;
            js_UseStepTargets.val = _jsonFile["MovementParameters"][10].AsBool;

            if (_jsonFile["MovementParameters"][11] != null)
                js_RotationStopAngle.val = _jsonFile["MovementParameters"][11].AsFloat;
        }

        private void GenerateShortcuts()
        {
            uFileBrowser.FileBrowser browser = SuperController.singleton.fileBrowserUI;
            browser.shortCuts = new List<ShortCut>();

            browser.shortCuts.Add(new ShortCut());
            browser.shortCuts[0].displayName = "Root";
            browser.shortCuts[0].path = "";

            browser.shortCuts.Add(new ShortCut());
            browser.shortCuts[1].displayName = "Addon Packages";
            browser.shortCuts[1].path = "AddonPackages";

            browser.shortCuts.Add(new ShortCut());
            browser.shortCuts[2].displayName = "Custom";
            browser.shortCuts[2].path = "Custom";

            browser.shortCuts.Add(new ShortCut());
            browser.shortCuts[3].displayName = "Saves";
            browser.shortCuts[3].path = "Saves";

            browser.shortCuts.Add(new ShortCut());
            browser.shortCuts[4].displayName = "CLS Folder";
            browser.shortCuts[4].path = MAIN_DIR;
        }

        /* Trigger Presets */
        public void Callback_Trigger_LoadFullTriggerPreset()
        {
            uFileBrowser.FileBrowser browser = SuperController.singleton.fileBrowserUI;

            GenerateShortcuts();

            browser.SetTitle("Load Trigger Preset");            
            browser.fileRemovePrefix = null;
            browser.hideExtension = false;
            browser.keepOpen = false;
            browser.fileFormat = EXT_FULLTRIGGER;
            browser.showDirs = true;
            browser.browseVarFilesAsDirectories = true;
            browser.SetTextEntry(false);
            browser.ActivateFileNameField();
            browser.clearCurrentPathOnHide = false;
            browser.Show(LoadTriggerFullTriggerPreset);
        }

        public void Callback_Trigger_SaveFullTriggerPreset()
        {
            uFileBrowser.FileBrowser browser = SuperController.singleton.fileBrowserUI;

            GenerateShortcuts();

            browser.SetTitle("Save Trigger Preset");
            browser.fileRemovePrefix = null;
            browser.hideExtension = false;
            browser.keepOpen = false;
            browser.fileFormat = EXT_FULLTRIGGER;
            browser.showDirs = true;
            browser.browseVarFilesAsDirectories = false;
            browser.SetTextEntry(true);
            browser.ActivateFileNameField();
            browser.clearCurrentPathOnHide = false;
            browser.fileEntryField.text = String.Format("{0}.{1}", "", EXT_FULLTRIGGER);
            browser.Show(SaveTriggerFullTriggerPreset);
        }

        public void SaveTriggerFullTriggerPreset(string _path)
        {
            if (_path == "")
                return;

            if (!_path.ToLower().EndsWith(EXT_FULLTRIGGER.ToLower()))
            {
                _path += "." + EXT_FULLTRIGGER;
            }

            JSONClass _saveJSON = new JSONClass();

            _saveJSON["Triggers"] = new JSONArray();

            _saveJSON["Triggers"][0] = trigger_OnIdleStart.GetJSON().AsObject;
            _saveJSON["Triggers"][1] = trigger_OnStepIdleStart.GetJSON().AsObject;
            _saveJSON["Triggers"][2] = trigger_OnMovementStart.GetJSON().AsObject;
            _saveJSON["Triggers"][3] = trigger_OnDestinationReached.GetJSON().AsObject;
            _saveJSON["Triggers"][4] = trigger_RightStep.GetJSON().AsObject;
            _saveJSON["Triggers"][5] = trigger_LeftStep.GetJSON().AsObject;
            _saveJSON["Triggers"][6] = trigger_BothSteps.GetJSON().AsObject;

            SaveJSON(_saveJSON, _path);
        }

        public void LoadTriggerFullTriggerPreset(string _path)
        {
            if (_path == "")
                return;

            JSONNode _jsonFile = LoadJSON(_path);

            trigger_OnIdleStart.RestoreFromJSON(_jsonFile["Triggers"][0].AsObject);
            trigger_OnStepIdleStart.RestoreFromJSON(_jsonFile["Triggers"][1].AsObject);
            trigger_OnMovementStart.RestoreFromJSON(_jsonFile["Triggers"][2].AsObject);
            trigger_OnDestinationReached.RestoreFromJSON(_jsonFile["Triggers"][3].AsObject);
            trigger_RightStep.RestoreFromJSON(_jsonFile["Triggers"][4].AsObject);
            trigger_LeftStep.RestoreFromJSON(_jsonFile["Triggers"][5].AsObject);
            trigger_BothSteps.RestoreFromJSON(_jsonFile["Triggers"][6].AsObject);
        }


        public void Callback_Trigger_LoadDestinationReachedPreset()
        {
            uFileBrowser.FileBrowser browser = SuperController.singleton.fileBrowserUI;

            GenerateShortcuts();

            browser.SetTitle("Load Trigger Preset");
            browser.fileRemovePrefix = null;
            browser.hideExtension = false;
            browser.keepOpen = false;
            browser.fileFormat = EXT_DEST_TRIGGER;
            browser.showDirs = true;
            browser.browseVarFilesAsDirectories = true;
            browser.SetTextEntry(false);
            browser.ActivateFileNameField();
            browser.clearCurrentPathOnHide = false;
            browser.Show(LoadTriggerDestinationReachedPreset);
        }

        public void Callback_Trigger_SaveDestinationReachedPreset()
        {
            uFileBrowser.FileBrowser browser = SuperController.singleton.fileBrowserUI;

            GenerateShortcuts();

            browser.SetTitle("Save Trigger Preset");
            browser.fileRemovePrefix = null;
            browser.hideExtension = false;
            browser.keepOpen = false;
            browser.fileFormat = EXT_DEST_TRIGGER;
            browser.showDirs = true;
            browser.browseVarFilesAsDirectories = false;
            browser.SetTextEntry(true);
            browser.ActivateFileNameField();
            browser.clearCurrentPathOnHide = false;
            browser.fileEntryField.text = String.Format("{0}.{1}", "", EXT_DEST_TRIGGER);
            browser.Show(SaveTriggerDestinationReachedPreset);
        }

        public void SaveTriggerDestinationReachedPreset(string _path)
        {
            if (_path == "")
                return;

            if (!_path.ToLower().EndsWith(EXT_DEST_TRIGGER.ToLower()))
            {
                _path += "." + EXT_DEST_TRIGGER;
            }

            JSONClass _saveJSON = new JSONClass();

            _saveJSON["Triggers"] = new JSONArray();

            _saveJSON["Triggers"][0] = trigger_OnDestinationReached.GetJSON().AsObject;

            SaveJSON(_saveJSON, _path);
        }

        public void LoadTriggerDestinationReachedPreset(string _path)
        {
            if (_path == "")
                return;

            JSONNode _jsonFile = LoadJSON(_path);

            trigger_OnDestinationReached.RestoreFromJSON(_jsonFile["Triggers"][0].AsObject);
        }

        /* Actions Callbacks */

        public void Callback_TriggerEditorChooser(string _choice)
        {
            switch (_choice)
            {
                case "OnIdleStart":
                    trigger_OnIdleStart.OpenPanelActionStart();
                    break;

                case "OnStepIdleStart":
                    trigger_OnStepIdleStart.OpenPanelActionStart();
                    break;

                case "OnMovementStart":
                    trigger_OnMovementStart.OpenPanelActionStart();
                    break;

                case "OnDestinationReached":
                    trigger_OnDestinationReached.OpenPanelActionStart();
                    break;
            }

            js_TriggerEditorChooser.val = triggerList[0];
        }

        public void Callback_TransitionTriggerEditorChooser(string _choice)
        {
            switch (_choice)
            {

                case "Right Step":
                    trigger_RightStep.OpenPanelTransition();
                    break;

                case "Left Step":
                    trigger_LeftStep.OpenPanelTransition();
                    break;

                case "Both Steps":
                    trigger_BothSteps.OpenPanelTransition();
                    break;
            }

            js_TransitionTriggerEditorChooser.val = transitionTriggerList[0];
        }

        public void Event_StartIdle(bool _newState)
        {
            if (_newState)
                trigger_OnIdleStart.Trigger();
        }

        public void Event_DestinationReached()
        {
            trigger_OnDestinationReached.Trigger();
        }

        /* Input Movement Callbacks */

        public void Callback_MovementInputReferenceChanged(string _newVal)
        {
            List<Atom> atoms = GetSceneAtoms();

            if (_newVal == "[None]")
            {
                movementRefTransform = null;
            }
            else if (_newVal == "[Player]")
            {
                movementRefTransform = CameraTarget.centerTarget.transform;
            }
            else
            {
                foreach (Atom a in atoms)
                {
                    if (a.name == _newVal)
                    {
                        movementRefTransform = a.mainController.transform;
                    }
                }
            }
        }

        public void Callback_MovementInputReferenceOpened()
        {
            js_MovementReferenceAtom.choices = GetAtomNames();
        }

        public void Callback_MovementInputToggled(bool _newVal)
        {
            if (positionTargetTransform == null)
                return;

            if(_newVal)
                positionTargetTransform.position = GetMainController().transform.position;
        }

        /* Helpers */

        public FreeControllerV3 GetControllerByName(string name)
        {
            FreeControllerV3[] ctrls = GetAllControllers();

            foreach(FreeControllerV3 ctrl in ctrls)
            {
                if (ctrl.name == name)
                    return ctrl;
            }

            return null;
        }

        public ForceReceiver GetForceReceiverByName(string name)
        {
            foreach(ForceReceiver _receiver in containingAtom.forceReceivers)
            {  
                if (_receiver.name == name)
                    return _receiver;             
            }

            return null;
        }

        public List<string> GetAtomNames()
        {
            List<Atom> atoms = GetSceneAtoms();
            List<string> atomsNames = new List<string>();

            atomsNames.Add("[None]");
            atomsNames.Add("[Player]");

            foreach (Atom a in atoms)
            {
                atomsNames.Add(a.name);
            }

            atomsNames.Sort();

            atomsNames.Remove("CoreControl");
            atomsNames.Remove("[CameraRig]");

            return atomsNames;
        }

        public static Vector3 StringToVector3(string _stringVector)
        {
            // Remove the parentheses
            if (_stringVector.StartsWith("(") && _stringVector.EndsWith(")"))
            {
                _stringVector = _stringVector.Substring(1, _stringVector.Length - 2);
            }

            // split the items
            string[] sArray = _stringVector.Split(',');

            // store as a Vector3
            Vector3 result = new Vector3(
                float.Parse(sArray[0]),
                float.Parse(sArray[1]),
                float.Parse(sArray[2]));

            return result;
        }

        private void OnAtomRename(string oldid, string newid)
        {
            foreach (EventTrigger et in eventTriggers)
            {
                et.SyncAtomNames();
            }
        }

        private static void DisableScrollOnText(UIDynamicTextField target)
        {
            ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
            if (targetSR != null)
            {
                targetSR.horizontal = false;
                targetSR.vertical = false;
            }
        }

        /* Step Target System */

        private List<MVRScript> stepTargets;

        protected void Callback_RefreshStepTargetList()
        {
            stepTargets = new List<MVRScript>();

            foreach (Atom _a in GetSceneAtoms())
            {
                foreach(MVRScript _script in _a.GetComponentsInChildren<MVRScript>())
                {
                    if(_script.name.Contains("CLS_Step"))
                    {
                        stepTargets.Add(_script);
                    }
                }
            }
        }

        /* Overrides */

        public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
        {
            JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
            if (includePhysical || forceStore)
            {
                needsStore = true;
                foreach (EventTrigger et in eventTriggers)
                {
                    jc[et.Name] = et.GetJSON(base.subScenePrefix);
                }

                foreach(EventTrigger et in transitionTriggers)
                {
                    jc[et.Name] = et.GetJSON(base.subScenePrefix);
                }

            }
            return jc;
        }

        public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
        {
            base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
            if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
            {
                foreach (EventTrigger et in eventTriggers)
                {
                    et.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
                }

                foreach (EventTrigger et in transitionTriggers)
                {
                    et.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
                }
            }
        }     
    }

    public class SimpleTriggerHandler : TriggerHandler
    {
        public static bool Loaded { get; private set; }

        private static SimpleTriggerHandler myInstance;

        private RectTransform myTriggerActionsPrefab;
        private RectTransform myTriggerActionMiniPrefab;
        private RectTransform myTriggerActionDiscretePrefab;
        private RectTransform myTriggerActionTransitionPrefab;

        public static SimpleTriggerHandler Instance
        {
            get
            {
                if (myInstance == null)
                    myInstance = new SimpleTriggerHandler();
                return myInstance;
            }
        }

        public static void LoadAssets()
        {
            SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal());
        }

        private IEnumerator LoadAssetsInternal()
        {
            foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", p => myTriggerActionsPrefab = p))
                yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", p => myTriggerActionMiniPrefab = p))
                yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", p => myTriggerActionDiscretePrefab = p))
                yield return x;
            foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", p => myTriggerActionTransitionPrefab = p))
                yield return x;

            Loaded = true;
        }

        private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
        {
            AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
            if (request == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
            yield return request;
            GameObject go = request.GetAsset<GameObject>();
            if (go == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
            RectTransform prefab = go.GetComponent<RectTransform>();
            if (prefab == null)
                throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
            assign(prefab);
        }


        void TriggerHandler.RemoveTrigger(Trigger t)
        {
            // nothing to do
        }

        void TriggerHandler.DuplicateTrigger(Trigger t)
        {
            throw new NotImplementedException();
        }

        RectTransform TriggerHandler.CreateTriggerActionsUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionMiniUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionDiscreteUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
        }

        RectTransform TriggerHandler.CreateTriggerActionTransitionUI()
        {
            return UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);
        }

        void TriggerHandler.RemoveTriggerActionUI(RectTransform rt)
        {
            UnityEngine.Object.Destroy(rt?.gameObject);
        }
    }

    // Wrapper for easier handling of custom triggers.
    public class EventTrigger : Trigger
    {
        public string Name
        {
            get { return name; }
            set { name = value; myNeedInit = true; }
        }

        public MVRScript Owner
        {
            get; private set;
        }

        private string name;
        private bool myNeedInit = true;

        public EventTrigger(MVRScript owner, string name)
        {
            Name = name;
            Owner = owner;
            handler = SimpleTriggerHandler.Instance;
        }

        public void OpenPanelActionStart()
        {
            if (!SimpleTriggerHandler.Loaded)
            {
                SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
                return;
            }

            triggerActionsParent = Owner.UITransform;
            InitTriggerUI();
            OpenTriggerActionsPanel();
            if (myNeedInit)
            {
                triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
                triggerActionsPanel.Find("Content/Tab1/Label").GetComponent<Text>().text = "Event Actions";
                triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(false);
                triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
                myNeedInit = false;
            }
        }

        public void OpenPanelTransition()
        {
            if (!SimpleTriggerHandler.Loaded)
            {
                SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
                return;
            }

            triggerActionsParent = Owner.UITransform;
            InitTriggerUI();
            OpenTriggerActionsPanel();

            //getRoot(triggerActionsParent,0);
            if (myNeedInit)
            {
                triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
                triggerActionsPanel.Find("Content/Tab2/Label").GetComponent<Text>().text = "Transition Actions";
                triggerActionsPanel.Find("Content/Tab1").gameObject.SetActive(false);
                triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(true);
                triggerActionsPanel.Find("Content/Tab2").GetComponent<Toggle>().isOn = true;
                triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
                myNeedInit = false;
            }
        }

        public void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge, bool setMissingToDefault)
        {
            if (jc.HasKey(Name))
            {
                JSONClass tc = jc[Name].AsObject;
                if (tc != null)
                    base.RestoreFromJSON(tc, subScenePrefix, isMerge);
            }
            else if (setMissingToDefault)
            {
                base.RestoreFromJSON(new JSONClass());
            }
        }

        public void Trigger()
        {
            active = true;
            active = false;
        }

        public void TriggerTransition(float val)
        {
            this.transitionInterpValue = val;

            foreach(TriggerActionTransition action in transitionActions)
            {
                action.TriggerInterp(val, true);
            }
        }
    }
}