﻿using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using MVR.FileManagementSecure;
using SimpleJSON;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

/* --- © 2022 JaxZoa <JaxZoa@gmail.com> --- */

namespace JaxPlugins
{
    public class CLS_CameraController : MVRScript
    {
        protected JSONStorableBool js_Active;
        protected JSONStorableBool js_InvertXAxis;
        protected JSONStorableBool js_InvertYAxis;

        protected JSONStorableBool js_CanZoom;
        protected JSONStorableFloat js_MinZoom;
        protected JSONStorableFloat js_MaxZoom;
        protected JSONStorableFloat js_ZoomSpeed;

        protected JSONStorableBool js_CanPanXY;
        protected JSONStorableBool js_InvertPanningX;
        protected JSONStorableBool js_InvertPanningY;
        protected JSONStorableFloat js_PanSpeed;

        protected JSONStorableStringChooser js_CameraAtom;
        protected JSONStorableStringChooser js_LookAtAtom;
        protected JSONStorableStringChooser js_LookAtController;
        protected JSONStorableStringChooser js_RotationAtom;
        protected JSONStorableStringChooser js_RotationController;

        protected JSONStorableStringChooser js_IgnoredCollidersAtom;

        protected JSONStorableBool js_AutoFollow;

        protected JSONStorableFloat js_LookAt_XOffset;
        protected JSONStorableFloat js_LookAt_YOffset;
        protected JSONStorableFloat js_LookAt_ZOffset;

        protected JSONStorableBool js_InputLockedRotationMode;
        protected JSONStorableFloat js_CameraMovementSpeed;
        protected JSONStorableFloat js_MaxCameraAngleUp;
        protected JSONStorableFloat js_MaxCameraAngleDown;

        protected JSONStorableFloat js_CameraSmoothing;

        protected JSONStorableFloat js_Camera_Distance;

        private Atom cameraAtom;
        private Atom lookAtAtom;
        private Atom ignoredCollidersAtom;
        private FreeControllerV3 lookAtController;
        private Atom rotationAtom;
        private FreeControllerV3 rotationController;

        private float xAxis, yAxis, zoomAxis;
        private Vector3 targetPos, currentPos;
        private float currentYaw, currentPitch;
        private Quaternion targetRot;

        private float currentCameraDistance;
        private float currentXOffset, currentYOffset, currentZOffset;

        private bool panningMode = false;
        private bool rotationMode = false;

        private Vector3 panningDirection;

        private List<Collider> targetColliders;
        private List<Collider> cameraColliders;
        private List<Collider> ignoredColliders;
        private float finalCameraDistance;
        RaycastHit[] hitResults;

        public override void Init()
        {
            hitResults = new RaycastHit[200];
            ignoredColliders = new List<Collider>();

            js_Active = new JSONStorableBool("IsActive", false);
            js_Active.setCallbackFunction += Callback_SetActive;

            js_InvertXAxis = new JSONStorableBool("Invert X Axis", false);
            js_InvertYAxis = new JSONStorableBool("Invert Y Axis", false);

            js_CanZoom = new JSONStorableBool("Can Zoom", true);
            js_ZoomSpeed = new JSONStorableFloat("Zoom Speed", 5f, 0f, 10f, false);
            js_MinZoom = new JSONStorableFloat("Min Zoom Distance", 0.3f, 0f, 2f, false);
            js_MaxZoom = new JSONStorableFloat("Max Zoom Distance", 5f, 0f, 5f, false);

            js_CanPanXY = new JSONStorableBool("Panning", true);
            js_InvertPanningX = new JSONStorableBool("Invert Panning X", true);
            js_InvertPanningY = new JSONStorableBool("Invert Panning Y", false);
            js_PanSpeed = new JSONStorableFloat("Panning Speed", 3f, 0f, 5f, false);

            js_CameraAtom = new JSONStorableStringChooser("CameraAtom", new List<string>(), "Select Atom", "Camera Atom", CameraAtomChanged);
            js_CameraAtom.popupOpenCallback += CameraAtomPopupOpened;

            js_LookAtAtom = new JSONStorableStringChooser("LookAtAtom", new List<string>(), "Select Atom", "Look At Atom", TargetAtomChanged);
            js_LookAtAtom.popupOpenCallback += TargetAtomPopupOpened;

            js_LookAtController = new JSONStorableStringChooser("LookAtController", new List<string>(), "Select Controller", "Look At Controller", TargetControllerChanged);

            js_RotationAtom = new JSONStorableStringChooser("RotationAtom", new List<string>(), "Select Atom", "Rotation Atom", RotationAtomChanged);
            js_RotationAtom.popupOpenCallback += RotationAtomPopupOpened;

            js_AutoFollow = new JSONStorableBool("Auto Follow", false);

            js_RotationController = new JSONStorableStringChooser("RotationController", new List<string>(), "Select Controller", "Rotation Controller", RotationControllerChanged);

            js_LookAt_XOffset = new JSONStorableFloat("Target X Offset", 0f, -1f, 1f, false);
            js_LookAt_YOffset = new JSONStorableFloat("Target Y Offset", 0f, -1f, 1f, false);
            js_LookAt_ZOffset = new JSONStorableFloat("Target Z Offset", 0f, -1f, 1f, false);


            js_InputLockedRotationMode = new JSONStorableBool("InputLocked Rotation", false);
            js_CameraMovementSpeed = new JSONStorableFloat("Camera Move Speed", 80f, 0.1f, 160f);
            js_MaxCameraAngleUp = new JSONStorableFloat("Camera Max Angle Up", 89f, 5f, 89f);
            js_MaxCameraAngleDown = new JSONStorableFloat("Camera Max Angle Down", 89f, 5f, 89f);

            js_Camera_Distance = new JSONStorableFloat("Camera Distance", 3.5f, 0f, 5f, false);

            js_CameraSmoothing = new JSONStorableFloat("Movement Smoothing", 0.2f, 0f, 1f);

            js_IgnoredCollidersAtom = new JSONStorableStringChooser("IgnoredCollidersAtom", new List<string>(), "Select Atom", "Ignored Colliders Atom", IgnoredCollidersAtomChanged);
            js_IgnoredCollidersAtom.popupOpenCallback += IgnoredCollidersAtomPopupOpened;


            RegisterBool(js_Active);

            RegisterStringChooser(js_CameraAtom);

            RegisterBool(js_InvertXAxis);
            RegisterBool(js_InvertYAxis);
            RegisterBool(js_AutoFollow);

            RegisterBool(js_CanZoom);
            RegisterFloat(js_MinZoom);
            RegisterFloat(js_MaxZoom);
            RegisterFloat(js_ZoomSpeed);

            RegisterBool(js_CanPanXY);
            RegisterBool(js_InvertPanningX);
            RegisterBool(js_InvertPanningY);
            RegisterFloat(js_PanSpeed);

            RegisterStringChooser(js_LookAtAtom);
            RegisterStringChooser(js_LookAtController);

            RegisterStringChooser(js_RotationAtom);
            RegisterStringChooser(js_RotationController);

            RegisterStringChooser(js_IgnoredCollidersAtom);

            RegisterFloat(js_LookAt_XOffset);
            RegisterFloat(js_LookAt_YOffset);
            RegisterFloat(js_LookAt_ZOffset);

            RegisterBool(js_InputLockedRotationMode);
            RegisterFloat(js_CameraMovementSpeed);
            RegisterFloat(js_MaxCameraAngleUp);
            RegisterFloat(js_MaxCameraAngleDown);

            RegisterFloat(js_Camera_Distance);

            RegisterFloat(js_CameraSmoothing);

            GenerateUI();
        }

        void Start()
        {
            Callback_SetActive(js_Active.val);

            ResetCamera();
        }

        protected void GenerateUI()
        {
            CreateToggle(js_Active);          
            
            CreateText("Title", "<color=#000><size=35><b>Required</b></size></color>", false, 50);

            CreateFilterablePopup(js_CameraAtom);
            CreateToggle(js_AutoFollow);

            CreateSpacer(25f);

            CreateFilterablePopup(js_LookAtAtom);
            CreateFilterablePopup(js_LookAtController);

            CreateSpacer(25f);

            CreateFilterablePopup(js_RotationAtom);
            CreateFilterablePopup(js_RotationController);

            CreateSpacer(25f);

            CreateFilterablePopup(js_IgnoredCollidersAtom);

            CreateText("Title", "<color=#000><size=35><b>Input</b></size></color>", false, 50);

            CreateToggle(js_InputLockedRotationMode);

            CreateToggle(js_InvertXAxis);
            CreateToggle(js_InvertYAxis);

            CreateSpacer(25f);

            CreateToggle(js_CanZoom);
            CreateSlider(js_ZoomSpeed);
            CreateSlider(js_MinZoom);
            CreateSlider(js_MaxZoom);

            CreateText("Title", "<color=#000><size=35><b>Placement</b></size></color>", true, 50);


            CreateSlider(js_Camera_Distance, true);
            CreateSlider(js_CameraSmoothing, true);

            CreateSlider(js_CameraMovementSpeed, true);
            CreateSlider(js_MaxCameraAngleUp, true);
            CreateSlider(js_MaxCameraAngleDown, true);

            CreateSpacer(25f, true);

            CreateSlider(js_LookAt_XOffset, true);
            CreateSlider(js_LookAt_YOffset, true);
            CreateSlider(js_LookAt_ZOffset, true);

            CreateText("Title", "<color=#000><size=35><b>Panning</b></size></color>", true, 50);

            CreateToggle(js_CanPanXY, true);
            CreateToggle(js_InvertPanningX, true);
            CreateToggle(js_InvertPanningY, true);
            CreateSlider(js_PanSpeed, true);
        }

        void Update()
        {
            if (!js_Active.val || lookAtController == null || cameraAtom == null || rotationController == null)
                return;

            UpdateInputs();
            UpdateTargetOffset();
            UpdateZoom();
            UpdateTargetPos();
            UpdateTargetRot();
            UpdateTargetDistance();
            
            currentPos = targetPos + (targetRot * Vector3.forward * finalCameraDistance);

            cameraAtom.mainController.transform.position = Vector3.Lerp(cameraAtom.mainController.transform.position, currentPos, 1 - js_CameraSmoothing.val);

            cameraAtom.mainController.transform.LookAt(targetPos);
        }

        void UpdateInputs()
        {
            panningMode = Input.GetMouseButton(2) || Input.GetKey(KeyCode.Joystick1Button4);
            rotationMode = Input.GetMouseButton(1) || Input.GetKey(KeyCode.Joystick1Button5);

            /* Keyboard & Mouse */

            xAxis = Input.GetAxis("Mouse X");
            yAxis = Input.GetAxis("Mouse Y");
            zoomAxis = -Input.mouseScrollDelta.y;
            zoomAxis += -Input.GetAxis(JoystickControl.singleton.triggersAxisName);


            if (Input.GetKeyDown(KeyCode.R) || Input.GetKey(KeyCode.Joystick1Button9))
            {
                currentXOffset = js_LookAt_XOffset.val;
                currentYOffset = js_LookAt_YOffset.val;
                currentZOffset = js_LookAt_ZOffset.val;
            }

            if(Input.GetKeyDown(KeyCode.Escape) )
            {
                js_Active.val = false;
            }

            /* Controllers */

            xAxis += Input.GetAxis(JoystickControl.singleton.rightStickXAxisName);
            yAxis += -Input.GetAxis(JoystickControl.singleton.rightStickYAxisName);
        }

        void UpdateTargetPos()
        {
            targetPos = lookAtController.transform.position;

            targetPos += rotationController.transform.forward * currentZOffset;
            targetPos += rotationController.transform.right * currentXOffset;
            targetPos += rotationController.transform.up * currentYOffset;
        }

        void UpdateTargetRot()
        {
            if ((rotationMode || !js_InputLockedRotationMode.val) && !panningMode)
            {
                float _xAxis = xAxis;
                float _yAxis = yAxis;

                if (js_InvertXAxis.val)
                    _xAxis = -_xAxis;

                if (js_InvertYAxis.val)
                    _yAxis = -_yAxis;

                currentYaw += _xAxis * Time.deltaTime * js_CameraMovementSpeed.val;

                if (currentYaw >= 360)
                    currentYaw = 0;
                else if (currentYaw < 0)
                    currentYaw = 360;

                currentPitch += _yAxis * Time.deltaTime * js_CameraMovementSpeed.val;

                currentPitch = Mathf.Clamp(currentPitch, -js_MaxCameraAngleDown.val, js_MaxCameraAngleUp.val);
            }

            targetRot = Quaternion.Euler(currentPitch, currentYaw, 0);

            if (js_AutoFollow.val)
                targetRot = rotationController.transform.rotation * targetRot;
        }

        void UpdateTargetOffset()
        {
            if(panningMode && js_CanPanXY.val)
            {
                float _axis = js_InvertPanningX.val ? -xAxis : xAxis;

                panningDirection = cameraAtom.mainController.transform.right * _axis * Time.deltaTime * js_PanSpeed.val;

                _axis = js_InvertPanningY.val ? -yAxis : yAxis;

                panningDirection += cameraAtom.mainController.transform.up * _axis * Time.deltaTime * js_PanSpeed.val;

                panningDirection = rotationController.transform.InverseTransformDirection(panningDirection);


                currentXOffset += panningDirection.x;
                currentYOffset += panningDirection.y;
                currentZOffset += panningDirection.z;
            }
        }

        void UpdateZoom()
        {
            if (js_CanZoom.val)
            {
                currentCameraDistance = Mathf.Clamp(currentCameraDistance + (zoomAxis * js_ZoomSpeed.val * Time.deltaTime), js_MinZoom.val, js_MaxZoom.val);
            }
            else
            {
                currentCameraDistance = js_Camera_Distance.val;
            }
        }


        float distance;
        float hitCount = 0;

        void UpdateTargetDistance()
        {
            distance = 1000;

            hitCount = Physics.SphereCastNonAlloc(targetPos, 0.01f, targetRot * Vector3.forward, hitResults, currentCameraDistance);

            if (hitCount > 0)
            {
                for(int i = 0; i < hitCount; i++)
                {
                    if (ignoredColliders.Contains(hitResults[i].collider))
                        continue;

                    if (targetColliders.Contains(hitResults[i].collider))
                        continue;

                    if (cameraColliders.Contains(hitResults[i].collider))
                        continue;

                    if((hitResults[i].point - targetPos).magnitude < distance)
                        distance = (hitResults[i].point - targetPos).magnitude;                   
                }
            }

            if (distance < currentCameraDistance)
            {
                finalCameraDistance = distance;
            }
            else
            {
                finalCameraDistance = currentCameraDistance;
            }
        }

        protected void ResetCamera()
        {
            currentCameraDistance = js_Camera_Distance.val;

            currentXOffset = js_LookAt_XOffset.val;
            currentYOffset = js_LookAt_YOffset.val;
            currentZOffset = js_LookAt_ZOffset.val;

            currentYaw = 0;
            currentPitch = 0;
        }

        protected void TargetAtomChanged(string _newVal)
        {
            List<Atom> atoms = GetSceneAtoms();

            lookAtController = null;

            if (_newVal == "[None]")
            {
                lookAtAtom = null;
            }
            else
            {
                foreach (Atom a in atoms)
                {
                    if (a.name == _newVal)
                    {
                        lookAtAtom = a;
                    }
                }
            }

            var _choices = new List<string>();

            if (lookAtAtom != null)
            { 
                foreach(FreeControllerV3 _fcv3 in lookAtAtom.freeControllers)
                {
                    _choices.Add(_fcv3.name);
                }

                js_LookAtController.choices = _choices;
            }

            if (lookAtAtom != null)
            {
                targetColliders = new List<Collider>();

                targetColliders.AddRange(lookAtAtom.GetComponentsInChildren<Collider>());
            }
        }

        protected void TargetAtomPopupOpened()
        {
            js_LookAtAtom.choices = GetAtomNames();
        }

        protected void TargetControllerChanged(string _newVal)
        {
            lookAtController = null;

            if (lookAtAtom == null)
                return;

            foreach (FreeControllerV3 _fcv3 in lookAtAtom.freeControllers)
            {
                if(_fcv3.name == _newVal)
                {
                    lookAtController = _fcv3;
                    return;
                }
            }
        }

        protected void RotationAtomChanged(string _newVal)
        {
            List<Atom> atoms = GetSceneAtoms();

            rotationController = null;

            if (_newVal == "[None]")
            {
                rotationAtom = null;
            }
            else
            {
                foreach (Atom a in atoms)
                {
                    if (a.name == _newVal)
                    {
                        rotationAtom = a;
                    }
                }
            }

            var _choices = new List<string>();

            if (rotationAtom != null)
            {
                foreach (FreeControllerV3 _fcv3 in rotationAtom.freeControllers)
                {
                    _choices.Add(_fcv3.name);
                }

                js_RotationController.choices = _choices;
            }
        }

        protected void RotationAtomPopupOpened()
        {
            js_RotationAtom.choices = GetAtomNames();
        }

        protected void RotationControllerChanged(string _newVal)
        {
            rotationController = null;

            if (rotationAtom == null)
                return;

            foreach (FreeControllerV3 _fcv3 in rotationAtom.freeControllers)
            {
                if (_fcv3.name == _newVal)
                {
                    rotationController = _fcv3;
                    return;
                }
            }
        }

        protected void CameraAtomChanged(string _newVal)
        {
            List<Atom> atoms = GetSceneAtoms();

            if (_newVal == "[None]")
            {
                cameraAtom = null;
            }
            else
            {
                foreach (Atom a in atoms)
                {
                    if (a.name == _newVal)
                    {
                        cameraAtom = a;
                    }
                }
            }

            if (cameraAtom != null)
            {
                cameraColliders = new List<Collider>();
                
                cameraColliders.AddRange(cameraAtom.GetComponentsInChildren<Collider>());
            }
        }

        protected void CameraAtomPopupOpened()
        {
            List<string> _names = GetAtomNames();
            js_CameraAtom.choices = _names;
        }

        protected void IgnoredCollidersAtomChanged(string _newVal)
        {
            List<Atom> atoms = GetSceneAtoms();
            ignoredColliders = new List<Collider>();

            if (_newVal == "[None]")
            {
                ignoredCollidersAtom = null;
            }
            else
            {
                foreach (Atom a in atoms)
                {
                    if (a.name == _newVal)
                    {
                        ignoredCollidersAtom = a;
                    }
                }
            }

            if (ignoredCollidersAtom != null)
            {
                ignoredColliders.AddRange(ignoredCollidersAtom.GetComponentsInChildren<Collider>());
            }
        }

        protected void IgnoredCollidersAtomPopupOpened()
        {
            js_IgnoredCollidersAtom.choices = GetAtomNames();
        }

        protected void Callback_SetActive(bool _newVal)
        {
            if (cameraAtom == null)
                return;

            CameraControl _cam = cameraAtom.GetComponentInChildren<CameraControl>();

            if (_newVal)
            {
                SuperController.singleton.MonitorCenterCamera.enabled = false;

                if (_cam != null)
                    _cam.SetBoolParamValue("cameraOn", true);
            }
            else
            {
                SuperController.singleton.MonitorCenterCamera.enabled = true;

                if (_cam != null)
                    _cam.SetBoolParamValue("cameraOn", false);
            }
        }

        public List<string> GetAtomNames()
        {
            List<Atom> atoms = GetSceneAtoms();
            List<string> atomsNames = new List<string>();

            atomsNames.Add("[None]");

            foreach (Atom a in atoms)
            {
                atomsNames.Add(a.name);
            }

            atomsNames.Sort();

            atomsNames.Remove("CoreControl");
            atomsNames.Remove("[CameraRig]");

            return atomsNames;
        }

        public void CreateSpacer(float _height, bool _right = false)
        {
            UIDynamic _spacer = CreateSpacer(_right);
            _spacer.height = _height;
        }

        public UIDynamicTextField CreateText(string DescTitle, string DescText, bool rightSide, int fieldHeight)
        {
            JSONStorableString staticDescString = new JSONStorableString(DescTitle, DescText);
            UIDynamicTextField staticDescStringField = CreateTextField(staticDescString, rightSide);
            staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
            LayoutElement sdsfLayout = staticDescStringField.GetComponent<LayoutElement>();
            sdsfLayout.preferredHeight = sdsfLayout.minHeight = fieldHeight;
            staticDescStringField.height = fieldHeight;
            DisableScrollOnText(staticDescStringField);

            return staticDescStringField;
        }

        private static void DisableScrollOnText(UIDynamicTextField target)
        {
            ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
            if (targetSR != null)
            {
                targetSR.horizontal = false;
                targetSR.vertical = false;
            }
        }
    }
}