﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;

/* --- © 2022 JaxZoa <JaxZoa@gmail.com> --- */

namespace JaxPlugins
{
    public class BodyController
    {
        protected CLS manager;

        protected StepCycle stepCycle;

        protected FreeControllerV3 mainControl, headControl, hipControl, chestControl, neckControl, rFootControl, lFootControl, rKneeControl, lKneeControl,rShoulderControl, lShoulderControl, rHandControl, lHandControl, rElbowControl, lElbowControl, rThighControl, lThighControl;
        protected ForceReceiver hipForceReceiver, chestForceReceiver, neckForceReceiver, rFootForceReceiver, lFootForceReceiver, rKneeForceReceiver, lKneeForceReceiver, rShoulderForceReceiver, lShoulderForceReceiver, rHandForceReceiver, lHandForceReceiver, rElbowForceReceiver, lElbowForceReceiver, rThighForceReceiver, lThighForceReceiver;

        protected Vector3 hipTargetPos, chestTargetPos, headTargetPos, rFootTargetPos, lFootTargetPos, rKneeTargetPos, lKneeTargetPos, rShoulderTargetPos, lShoulderTargetPos, rHandTargetPos, lHandTargetPos, rElbowTargetPos, lElbowTargetPos;
        protected Quaternion hipTargetRot, chestTargetRot, headTargetRot, rFootTargetRot, lFootTargetRot, rHandTargetRot, lHandTargetRot;

        protected bool updateHipTarget = true;
        protected bool updateChestTarget = true;
        protected bool updateHeadTarget = true;
        protected bool updateRFootTarget = true;
        protected bool updateLFootTarget = true;
        protected bool updateRKneeTarget = true;
        protected bool updateLkneeTarget = true;
        protected bool updateRShoulderTarget = true;
        protected bool updateLShoulderTarget = true;
        protected bool updateRHandTarget = true;
        protected bool updateLHandTarget = true;
        protected bool updateRElbowTarget = true;
        protected bool updateLElbowTarget = true;

        protected float controlsFollowTargets = 0.3f;

        protected JSONStorableBool js_StopControlingWhenIdle;        

        public BodyController(CLS _manager)
        {
            manager = _manager;

            mainControl = manager.GetControllerByName("control");
            headControl = manager.GetControllerByName("headControl");
            hipControl = manager.GetControllerByName("hipControl");
            chestControl = manager.GetControllerByName("chestControl");
            neckControl = manager.GetControllerByName("neckControl");
            rFootControl = manager.GetControllerByName("rFootControl");
            lFootControl = manager.GetControllerByName("lFootControl");
            rKneeControl = manager.GetControllerByName("rKneeControl");
            lKneeControl = manager.GetControllerByName("lKneeControl");
            rHandControl = manager.GetControllerByName("rHandControl");
            lHandControl = manager.GetControllerByName("lHandControl");
            rElbowControl = manager.GetControllerByName("rElbowControl");
            lElbowControl = manager.GetControllerByName("lElbowControl");
            rThighControl = manager.GetControllerByName("rThighControl");
            lThighControl = manager.GetControllerByName("lThighControl");
            lShoulderControl = manager.GetControllerByName("lArmControl");
            rShoulderControl = manager.GetControllerByName("rArmControl");

            hipForceReceiver= manager.GetForceReceiverByName("hip");
            chestForceReceiver = manager.GetForceReceiverByName("chest");
            neckForceReceiver = manager.GetForceReceiverByName("neck");
            rFootForceReceiver = manager.GetForceReceiverByName("rFoot");
            lFootForceReceiver = manager.GetForceReceiverByName("lFoot");
            rKneeForceReceiver = manager.GetForceReceiverByName("rShin");
            lKneeForceReceiver = manager.GetForceReceiverByName("lShin");
            rHandForceReceiver = manager.GetForceReceiverByName("rHand");
            lHandForceReceiver = manager.GetForceReceiverByName("lHand");
            rElbowForceReceiver = manager.GetForceReceiverByName("rForeArm");
            lElbowForceReceiver = manager.GetForceReceiverByName("lForeArm");
            rThighForceReceiver = manager.GetForceReceiverByName("rThigh");
            lThighForceReceiver = manager.GetForceReceiverByName("lThigh");
            lShoulderForceReceiver = manager.GetForceReceiverByName("lShldr");
            rShoulderForceReceiver = manager.GetForceReceiverByName("rShldr");

            js_StopControlingWhenIdle = new JSONStorableBool("Stop Control When Idle", false);
            manager.RegisterBool(js_StopControlingWhenIdle);
        }

        public virtual void Start()
        {
            
        }

        public virtual void Update()
        {
            stepCycle.ComputeSteps();
            UpdateTargets();
            UpdateBodyControls();
        }

        protected void UpdateBodyControls()
        {
            if (stepCycle.IsIdle && js_StopControlingWhenIdle.val)
                return;

            if (updateHipTarget)
            {
                hipControl.transform.position = Vector3.Lerp(hipControl.transform.position, hipTargetPos, controlsFollowTargets);
                hipControl.transform.rotation = Quaternion.Lerp(hipControl.transform.rotation, hipTargetRot, controlsFollowTargets);
            }

            if (updateChestTarget)
            {
                chestControl.transform.position = Vector3.Lerp(chestControl.transform.position, chestTargetPos, controlsFollowTargets);
                chestControl.transform.rotation = Quaternion.Lerp(chestControl.transform.rotation, chestTargetRot, controlsFollowTargets); ;
            }

            if (updateHeadTarget)
            {
                headControl.transform.position = Vector3.Lerp(headControl.transform.position, headTargetPos, controlsFollowTargets);
                headControl.transform.rotation = Quaternion.Lerp(headControl.transform.rotation, headTargetRot, controlsFollowTargets);
            }

            if (updateRFootTarget)
            {
                rFootControl.transform.position = Vector3.Lerp(rFootControl.transform.position, rFootTargetPos, controlsFollowTargets);
                rFootControl.transform.rotation = Quaternion.Lerp(rFootControl.transform.rotation, rFootTargetRot, controlsFollowTargets);
            }

            if (updateLFootTarget)
            {
                lFootControl.transform.position = Vector3.Lerp(lFootControl.transform.position, lFootTargetPos, controlsFollowTargets);
                lFootControl.transform.rotation = Quaternion.Lerp(lFootControl.transform.rotation, lFootTargetRot, controlsFollowTargets);
            }

            if (updateRKneeTarget)
                rKneeControl.transform.position = Vector3.Lerp(rKneeControl.transform.position, rKneeTargetPos, controlsFollowTargets);

            if (updateLkneeTarget)
                lKneeControl.transform.position = Vector3.Lerp(lKneeControl.transform.position, lKneeTargetPos, controlsFollowTargets);

            if (updateRHandTarget)
            {
                rHandControl.transform.position = Vector3.Lerp(rHandControl.transform.position, rHandTargetPos, controlsFollowTargets);
                rHandControl.transform.rotation = Quaternion.Lerp(rHandControl.transform.rotation, rHandTargetRot, controlsFollowTargets);
            }

            if (updateLHandTarget)
            {
                lHandControl.transform.position = Vector3.Lerp(lHandControl.transform.position, lHandTargetPos, controlsFollowTargets);
                lHandControl.transform.rotation = Quaternion.Lerp(lHandControl.transform.rotation, lHandTargetRot, controlsFollowTargets);
            }

            if (updateRElbowTarget)
                rElbowControl.transform.position = Vector3.Lerp(rElbowControl.transform.position, rElbowTargetPos, controlsFollowTargets);

            if (updateLElbowTarget)
                lElbowControl.transform.position = Vector3.Lerp(lElbowControl.transform.position, lElbowTargetPos, controlsFollowTargets);

            if (updateRShoulderTarget)
                rShoulderControl.transform.position = Vector3.Lerp(rShoulderControl.transform.position, rShoulderTargetPos, controlsFollowTargets);

            if (updateLShoulderTarget)
                lShoulderControl.transform.position = Vector3.Lerp(lShoulderControl.transform.position, lShoulderTargetPos, controlsFollowTargets);
        }

        public virtual void OnDestroy()
        {

        }

        protected virtual void UpdateTargets()
        {
            
        }

        public virtual void CreateGlobalUI()
        {
            
        }

        protected virtual void SetDefaultControllersState()
        {
            foreach (FreeControllerV3 controller in manager.GetAllControllers())
            {
                controller.currentPositionState = FreeControllerV3.PositionState.Off;
                controller.currentRotationState = FreeControllerV3.RotationState.Off;


                if (controller.name == "control")
                {
                    controller.currentPositionState = FreeControllerV3.PositionState.On;
                    controller.currentRotationState = FreeControllerV3.RotationState.On;
                    continue;
                }

                if (controller.name == "eyeTargetControl")
                {
                    controller.currentPositionState = FreeControllerV3.PositionState.On;
                    controller.currentRotationState = FreeControllerV3.RotationState.On;
                    continue;
                }

                if (controller.name == "headControl")
                {
                    controller.currentPositionState = FreeControllerV3.PositionState.On;
                    controller.currentRotationState = FreeControllerV3.RotationState.On;
                    continue;
                }

                if (controller.name == "hipControl")
                {
                    controller.currentPositionState = FreeControllerV3.PositionState.On;
                    controller.currentRotationState = FreeControllerV3.RotationState.On;
                    continue;
                }

                if (controller.name == "chestControl")
                {
                    controller.currentPositionState = FreeControllerV3.PositionState.On;
                    controller.currentRotationState = FreeControllerV3.RotationState.On;
                    continue;
                }

                if (controller.name == "lKneeControl")
                {
                    controller.currentPositionState = FreeControllerV3.PositionState.On;
                    continue;
                }

                if (controller.name == "rKneeControl")
                {
                    controller.currentPositionState = FreeControllerV3.PositionState.On;
                    continue;
                }

                if (controller.name == "lFootControl")
                {
                    controller.currentPositionState = FreeControllerV3.PositionState.On;
                    controller.currentRotationState = FreeControllerV3.RotationState.On;
                    continue;
                }

                if (controller.name == "rFootControl")
                {
                    controller.currentPositionState = FreeControllerV3.PositionState.On;
                    controller.currentRotationState = FreeControllerV3.RotationState.On;
                    continue;
                }

                if (controller.name == "lArmControl")
                {
                    controller.currentPositionState = FreeControllerV3.PositionState.On;
                    continue;
                }

                if (controller.name == "rArmControl")
                {
                    controller.currentPositionState = FreeControllerV3.PositionState.On;
                    continue;
                }

                if (controller.name == "lElbowControl")
                {
                    controller.currentPositionState = FreeControllerV3.PositionState.On;
                    continue;
                }

                if (controller.name == "rElbowControl")
                {
                    controller.currentPositionState = FreeControllerV3.PositionState.On;
                    continue;
                }

                if (controller.name == "lHandControl")
                {
                    controller.currentPositionState = FreeControllerV3.PositionState.On;
                    controller.currentRotationState = FreeControllerV3.RotationState.On;
                    continue;
                }

                if (controller.name == "rHandControl")
                {
                    controller.currentPositionState = FreeControllerV3.PositionState.On;
                    controller.currentRotationState = FreeControllerV3.RotationState.On;
                    continue;
                }

            }
        }

        public virtual JSONClass GetAllBodyParameters()
        {
            return new JSONClass();
        }

        public virtual void SetAllParametersFromSaveFile(JSONNode _saveJSON)
        {

        }

        public StepCycle StepCycle
        {
            get
            {
                return stepCycle;
            }
        }

        public CLS GetManager()
        {
            return manager;
        }

        public Vector3 CurrentBodyPos
        {
            get
            {
                return hipControl.transform.position;
            }
        }

        public Quaternion CurrentBodyRot
        {
            get
            {
                return Quaternion.Euler(0, hipControl.transform.rotation.eulerAngles.y, 0);
            }
        }

        public Vector3 CurrentFeetPos(float _leftLerp)
        {
            return Vector3.Lerp(rFootControl.transform.position, lFootControl.transform.position, _leftLerp);
        }

        public virtual bool IsStanding
        {
            get
            {
                return true;
            }
        }

        public virtual void BodyControllerMovementUI()
        {

        }

        public virtual void BodyControllerTriggersUI()
        {

        }

        public virtual void BodyControllerPresetsUI()
        {

        }
    }
}